"""Model refueling at several gas stations.

Each gas station has several fuel pumps and a single, shared reservoir. Each
arrving car pumps gas from the reservoir via a fuel pump.

As the gas station's reservoir empties, a request is made to a tanker truck
company to send a truck to refill the reservoir. The tanker company maintains a
fleet of tanker trucks.

This example demonstrates core desmod concepts including:
 - Modeling using Component subclasses
 - The "batteries-included" simulation environment
 - Centralized configuration
 - Logging

"""
from itertools import count, cycle

from simpy import Resource

from desmod.component import Component
from desmod.dot import generate_dot
from desmod.pool import Pool
from desmod.queue import Queue
from desmod.simulation import simulate


class Top(Component):
    """Every model has a single top-level Component.

    For this gas station model, the top level components are gas stations and a
    tanker truck company.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # The simulation configuration is available everywhere via the
        # simulation environment.
        num_gas_stations = self.env.config.get('gas_station.count', 1)

        # Instantiate GasStation components. An index is passed so that each
        # child gas station gets a unique name.
        self.gas_stations = [GasStation(self, index=i) for i in range(num_gas_stations)]

        # There is just one tanker company.
        self.tanker_company = TankerCompany(self)

    def connect_children(self):
        # This function is called during the elaboration phase, i.e. after all
        # of the components have been instantiated, but before the simulation
        # phase.
        for gas_station in self.gas_stations:
            # Each GasStation instance gets a reference to (is connected to)
            # the tanker_company instance. This demonstrates the most
            # abbreviated way to call connect().
            self.connect(gas_station, 'tanker_company')

    def elab_hook(self):
        generate_dot(self)


class TankerCompany(Component):
    """The tanker company owns and dispatches its fleet of tanker trunks."""

    # This base_name is used to build names and scopes of component instances.
    base_name = 'tankerco'

    def __init__(self, *args, **kwargs):
        # Many Component subclasses can simply forward *args and **kwargs to
        # the superclass initializer; although Component subclasses may also
        # have custom positional and keyword arguments.
        super().__init__(*args, **kwargs)
        num_tankers = self.env.config.get('tanker.count', 1)

        # Instantiate the fleet of tanker trucks.
        trucks = [TankerTruck(self, index=i) for i in range(num_tankers)]

        # Trucks are dispatched in a simple round-robin fashion.
        self.trucks_round_robin = cycle(trucks)

    def request_truck(self, gas_station, done_event):
        """Called by gas stations to request a truck to refill its reservior.

        Returns an event that the gas station must yield for.

        """
        truck = next(self.trucks_round_robin)

        # Each component has debug(), info(), warn(), and error() log methods.
        # Log lines are automatically annotated with the simulation time and
        # the scope of the component doing the logging.
        self.info(f'dispatching {truck.name} to {gas_station.name}')
        return truck.dispatch(gas_station, done_event)


class TankerTruck(Component):
    """Tanker trucks carry fuel to gas stations.

    Each tanker truck has a queue of gas stations it must visit. When the
    truck's tank becomes empty, it must go refill itself.

    """

    base_name = 'truck'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pump_rate = self.env.config.get('tanker.pump_rate', 10)
        self.avg_travel = self.env.config.get('tanker.travel_time', 600)
        tank_capacity = self.env.config.get('tanker.capacity', 200)
        self.tank = Pool(self.env, tank_capacity)

        # This auto_probe() call uses the self.tank Pool get/put hooks so that
        # whenever it's level changes, the new level is noted in the log.
        self.auto_probe('tank', log={})

        # The parent TankerCompany enqueues instructions to this queue.
        self._instructions = Queue(self.env)

        # Declare a persistant process to be started at simulation-time.
        self.add_process(self._dispatch_loop)

    def dispatch(self, gas_station, done_event):
        """Append dispatch instructions to the truck's queue."""
        return self._instructions.put((gas_station, done_event))

    def _dispatch_loop(self):
        """This is the tanker truck's main behavior. Travel, pump, refill..."""
        while True:
            if not self.tank.level:
                self.info('going for refill')

                # Desmod simulation environments come equipped with a
                # random.Random() instance seeded based on the 'sim.seed'
                # configuration key.
                travel_time = self.env.rand.expovariate(1 / self.avg_travel)
                yield self.env.timeout(travel_time)

                self.info('refilling')
                pump_time = self.tank.capacity / self.pump_rate
                yield self.env.timeout(pump_time)

                yield self.tank.put(self.tank.capacity)
                self.info(f'refilled {self.tank.capacity}L in {pump_time:.0f}s')

            gas_station, done_event = yield self._instructions.get()
            self.info(f'traveling to {gas_station.name}')
            travel_time = self.env.rand.expovariate(1 / self.avg_travel)
            yield self.env.timeout(travel_time)
            self.info(f'arrived at {gas_station.name}')
            while self.tank.level and (
                gas_station.reservoir.level < gas_station.reservoir.capacity
            ):
                yield self.env.timeout(1 / self.pump_rate)
                yield gas_station.reservoir.put(1)
                yield self.tank.get(1)
            self.info('done pumping')
            done_event.succeed()


class GasStation(Component):
    """A gas station has a fuel reservoir shared among several fuel pumps.

    The gas station has a traffic generator process that causes cars to arrive
    to fill up their tanks.

    As the cars fill up, the reservoir's level goes down. When the level goes
    below a critical threshold, the gas station makes a request to the tanker
    company for a tanker truck to refill the reservoir.

    """

    base_name = 'station'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        config = self.env.config
        self.add_connections('tanker_company')
        self.arrival_interval = config.get('gas_station.arrival_interval', 60)

        station_capacity = config.get('gas_station.capacity', 200)
        self.reservoir = Pool(
            self.env, capacity=station_capacity, init=station_capacity
        )
        self.auto_probe('reservoir', log={})

        threshold_pct = config.get('gas_station.threshold_pct', 10)
        self.reservoir_low_water = threshold_pct * station_capacity / 100

        self.pump_rate = config.get('gas_station.pump_rate', 2)
        num_pumps = config.get('gas_station.pumps', 2)
        self.fuel_pumps = Resource(self.env, capacity=num_pumps)
        self.auto_probe('fuel_pumps', log={})

        self.car_capacity = config.get('car.capacity', 50)
        self.car_level_range = config.get('car.level', [5, 25])

        # A gas station has two persistent processes. One to monitor the
        # reservoir level and one that models the arrival of cars at the
        # station. Desmod starts these processes before simulation phase.
        self.add_processes(self._monitor_reservoir, self._traffic_generator)

    @property
    def reservoir_pct(self):
        return self.reservoir.level / self.reservoir.capacity * 100

    def _monitor_reservoir(self):
        """Periodically monitor reservoir level.

        The a request is made to the tanker company when the reservoir falls
        below a critical threshold.

        """
        while True:
            yield self.reservoir.when_at_most(self.reservoir_low_water)
            done_event = self.env.event()
            yield self.tanker_company.request_truck(self, done_event)
            yield done_event

    def _traffic_generator(self):
        """Model the sporadic arrival of cars to the gas station."""
        for i in count():
            interval = self.env.rand.expovariate(1 / self.arrival_interval)
            yield self.env.timeout(interval)
            self.env.process(self._car(i))

    def _car(self, i):
        """Model a car transacting fuel."""
        with self.fuel_pumps.request() as pump_req:
            self.info(f'car{i} awaiting pump')
            yield pump_req
            self.info(f'car{i} at pump')
            car_level = self.env.rand.randint(*self.car_level_range)
            amount = self.car_capacity - car_level
            t0 = self.env.now
            for _ in range(amount):
                yield self.reservoir.get(1)
                yield self.env.timeout(1 / self.pump_rate)
            pump_time = self.env.now - t0
            self.info(f'car{i} pumped {amount}L in {pump_time:.0f}s')


# Desmod uses a plain dictionary to represent the simulation configuration.
# The various 'sim.xxx' keys are reserved for desmod while the remainder are
# application-specific.
config = {
    'car.capacity': 50,
    'car.level': [5, 25],
    'gas_station.capacity': 200,
    'gas_station.count': 3,
    'gas_station.pump_rate': 2,
    'gas_station.pumps': 2,
    'gas_station.arrival_interval': 60,
    'sim.dot.enable': True,
    'sim.dot.colorscheme': 'blues5',
    'sim.duration': '500 s',
    'sim.log.enable': True,
    'sim.log.file': 'sim.log',
    'sim.log.format': '{level:7} {ts:.3f} {ts_unit}: {scope:<16}:',
    'sim.log.level': 'INFO',
    'sim.result.file': 'results.yaml',
    'sim.seed': 42,
    'sim.timescale': 's',
    'sim.workspace': 'workspace',
    'tanker.capacity': 200,
    'tanker.count': 2,
    'tanker.pump_rate': 10,
    'tanker.travel_time': 100,
}

if __name__ == '__main__':
    # Desmod takes responsibility for instantiating and elaborating the model,
    # thus we only need to pass the configuration dict and the top-level
    # Component class (Top) to simulate().
    simulate(config, Top)



import string

_formatter = string.Formatter()


def partial_format(format_string: str, **kwargs: object) -> str:
    """Partially replace replacement fields in format string.

    Partial formatting allows a format string to be progressively formatted.
    This may be helpful to either amortize the expense of formatting or allow
    different entities (with access to different information) cooperatively
    format a string.

    Only named replacement fields are supported; positional replacement fields
    are not supported.

    :param str format_string: Format string to partially apply replacements.
    :param kwargs: Replacements for named fields.
    :returns: Partially formatted format string.

    """
    result = []
    for literal, field, spec, conversion in _formatter.parse(format_string):
        if literal:
            result.append(literal)
        if field is not None:
            spec_list = [field]
            if conversion:
                spec_list.extend(['!', conversion])
            if spec:
                spec_list.extend([':', spec])
            inner_spec = ''.join(spec_list)
            formatted_inner = partial_format(inner_spec, **kwargs)
            if not field or field.isdigit() or field not in kwargs:
                result.extend(['{{', formatted_inner, '}}'])
            else:
                result.extend(['{', formatted_inner, '}'])
    return ''.join(result).format(**kwargs)



import json
import os

import pytest
import yaml

from desmod.component import Component
from desmod.simulation import (
    SimEnvironment,
    SimStopEvent,
    simulate,
    simulate_factors,
    simulate_many,
)
import desmod.progress

pytestmark = pytest.mark.usefixtures('cleandir')


@pytest.fixture
def cleandir(tmpdir):
    origin = os.getcwd()
    tmpdir.chdir()
    yield None
    os.chdir(origin)


@pytest.fixture
def no_progressbar(monkeypatch):
    monkeypatch.setattr(desmod.progress, 'progressbar', None)


@pytest.fixture
def no_colorama(monkeypatch):
    monkeypatch.setattr(desmod.progress, 'colorama', None)


@pytest.fixture
def config():
    return {
        'sim.config.file': 'config.yaml',
        'sim.result.file': 'result.yaml',
        'sim.workspace': 'workspace',
        'sim.workspace.overwrite': False,
        'sim.timescale': '1 us',
        'sim.seed': 1234,
        'sim.duration': '1 us',
        'test.ensure_workspace': False,
        'test.fail_pre_init': False,
        'test.fail_init': False,
        'test.fail_simulate': False,
        'test.fail_post_simulate': False,
        'test.fail_get_result': False,
        'test.until_delay': None,
    }


class TopTest(Component):
    @classmethod
    def pre_init(cls, env):
        if env.config.get('test.fail_pre_init'):
            raise Exception('fail_pre_init')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.env.config.get('test.fail_init'):
            raise Exception('fail_init')
        self.add_process(self.test_proc)

    def test_proc(self):
        yield self.env.timeout(0.5)
        if self.env.config.get('test.fail_simulate'):
            assert False, 'fail_simulate'
        until_delay = self.env.config.get('test.until_delay')
        if until_delay is not None:
            self.env.until.schedule(until_delay)
        yield self.env.timeout(0.5)

    def post_sim_hook(self):
        if self.env.config.get('test.fail_post_simulate'):
            raise Exception('fail_post_simulate')

    def get_result_hook(self, result):
        if self.env.config.get('test.fail_get_result'):
            raise Exception('fail_get_result')
        result['time_100ps'] = self.env.time(unit='100 ps')


def test_pre_init_failure(config):
    config['test.fail_pre_init'] = True
    result = simulate(config, TopTest, reraise=False)
    assert result['sim.exception'] == repr(Exception('fail_pre_init'))
    assert result['sim.now'] == 0
    assert result['sim.time'] == 0
    assert result['sim.runtime'] > 0
    assert result['config']['test.fail_pre_init']
    assert os.path.exists(
        os.path.join(config['sim.workspace'], config['sim.result.file'])
    )


def test_init_failure(config):
    config['test.fail_init'] = True
    result = simulate(config, TopTest, reraise=False)
    assert result['sim.exception'] == repr(Exception('fail_init'))
    assert result['sim.now'] == 0
    assert result['sim.time'] == 0
    assert result['sim.runtime'] > 0
    assert result['config']['test.fail_init']
    for file_key in ['sim.result.file', 'sim.config.file']:
        assert os.path.exists(os.path.join(config['sim.workspace'], config[file_key]))


def test_simulate_fail(config):
    config['test.fail_simulate'] = True
    result = simulate(config, TopTest, reraise=False)
    assert result['sim.exception'].startswith('AssertionError')
    assert result['sim.now'] == 0.5
    assert result['sim.time'] == 0.5e-6
    assert result['sim.runtime'] > 0
    assert result['config']['test.fail_simulate']
    for file_key in ['sim.result.file', 'sim.config.file']:
        assert os.path.exists(os.path.join(config['sim.workspace'], config[file_key]))


def test_post_simulate_fail(config):
    config['test.fail_post_simulate'] = True
    result = simulate(config, TopTest, reraise=False)
    assert result['sim.exception'] == repr(Exception('fail_post_simulate'))
    assert result['sim.now'] == 1
    assert result['sim.time'] == 1e-6
    assert result['sim.runtime'] > 0
    assert result['config']['test.fail_post_simulate']
    for file_key in ['sim.result.file', 'sim.config.file']:
        assert os.path.exists(os.path.join(config['sim.workspace'], config[file_key]))


def test_get_result_fail(config):
    config['test.fail_get_result'] = True
    result = simulate(config, TopTest, reraise=False)
    assert result['sim.exception'] == repr(Exception('fail_get_result'))
    assert result['sim.now'] == 1
    assert result['sim.time'] == 1e-6
    assert result['sim.runtime'] > 0
    assert result['config']['test.fail_get_result']
    for file_key in ['sim.result.file', 'sim.config.file']:
        assert os.path.exists(os.path.join(config['sim.workspace'], config[file_key]))


def test_simulate_reraise(config):
    config['test.fail_simulate'] = True
    with pytest.raises(AssertionError):
        simulate(config, TopTest, reraise=True)


def test_no_result_file(config):
    config.pop('sim.result.file')
    config.pop('sim.config.file')
    result = simulate(config, TopTest)
    assert result['sim.exception'] is None
    assert not os.listdir(config['sim.workspace'])


def test_simulate_with_progress(config, capsys):
    config['sim.progress.enable'] = True
    config['sim.duration'] = '10 us'
    simulate(config, TopTest)
    _, err = capsys.readouterr()
    assert err.endswith('(100%)\n')


@pytest.mark.parametrize('max_width', [0, 1])
def test_simulate_with_progress_tty(config, capsys, max_width):
    config['sim.progress.enable'] = True
    config['sim.progress.max_width'] = max_width
    config['sim.duration'] = '10 us'
    with capsys.disabled():
        simulate(config, TopTest)


def test_simulate_progress_non_one_timescale(config):
    config['sim.progress.enable'] = True
    config['sim.timescale'] = '100 ns'
    config['sim.duration'] = '10 us'
    simulate(config, TopTest)


def test_simulate_factors(config):
    factors = [(['sim.seed'], [[1], [2], [3]])]
    results = simulate_factors(config, factors, TopTest)
    assert len(results) == 3
    for result in results:
        assert result['sim.exception'] is None
        assert os.path.exists(
            os.path.join(
                result['config']['meta.sim.workspace'],
                result['config']['sim.result.file'],
            )
        )


def test_simulate_factors_only_factor(config):
    FACTOR_NUM = 2

    def single_factor_filter_fn(cfg):
        return cfg['meta.sim.index'] == FACTOR_NUM

    factors = [(['sim.seed'], [[1], [2], [3]])]
    results = simulate_factors(
        config, factors, TopTest, config_filter=single_factor_filter_fn
    )
    assert len(results) == 1
    for result in results:
        assert result['sim.exception'] is None
        assert result['config']['meta.sim.workspace'] == os.path.join(
            config['sim.workspace'], str(FACTOR_NUM)
        )
        assert os.path.exists(
            os.path.join(
                result['config']['meta.sim.workspace'],
                result['config']['sim.result.file'],
            )
        )


def test_simulate_factors_progress(config, capfd):
    config['sim.progress.enable'] = True
    config['sim.duration'] = '10 us'
    factors = [(['sim.seed'], [[1], [2], [3]])]
    results = simulate_factors(config, factors, TopTest)
    assert len(results) == 3
    for result in results:
        assert result['sim.exception'] is None
        assert os.path.exists(
            os.path.join(
                result['config']['meta.sim.workspace'],
                result['config']['sim.result.file'],
            )
        )
    out, err = capfd.readouterr()
    assert out == ''
    assert '3 of 3 simulations' in err


def test_simulate_factors_progress_tty(config, capsys):
    config['sim.progress.enable'] = True
    config['sim.duration'] = '10 us'
    factors = [(['sim.seed'], [[1], [2], [3]])]
    with capsys.disabled():
        results = simulate_factors(config, factors, TopTest)
    assert len(results) == 3
    for result in results:
        assert result['sim.exception'] is None
        assert os.path.exists(
            os.path.join(
                result['config']['meta.sim.workspace'],
                result['config']['sim.result.file'],
            )
        )


def test_simulate_factors_no_overwrite(config):
    config['sim.workspace.overwrite'] = False
    factors = [(['sim.seed'], [[1], [2], [3]])]
    results = simulate_factors(config, factors, TopTest)
    assert os.path.isdir(config['sim.workspace'])
    assert len(results) == 3
    for result in results:
        assert result['sim.exception'] is None
        assert os.path.exists(
            os.path.join(
                result['config']['meta.sim.workspace'],
                result['config']['sim.result.file'],
            )
        )

    with open(os.path.join(config['sim.workspace'], 'cookie.txt'), 'w') as f:
        f.write('hi')

    factors = [(['sim.seed'], [[1], [2], [3], [4]])]
    results = simulate_factors(config, factors, TopTest)
    assert len(results) == 4
    assert os.path.isdir(config['sim.workspace'])
    for result in results:
        assert result['sim.exception'] is None
        assert os.path.exists(
            os.path.join(
                result['config']['meta.sim.workspace'],
                result['config']['sim.result.file'],
            )
        )

    with open(os.path.join(config['sim.workspace'], 'cookie.txt')) as f:
        assert f.read() == 'hi'


def test_simulate_factors_overwrite(config):
    config['sim.workspace.overwrite'] = True
    factors = [(['sim.seed'], [[1], [2], [3]])]
    results = simulate_factors(config, factors, TopTest)
    assert os.path.isdir(config['sim.workspace'])
    assert len(results) == 3
    for result in results:
        assert result['sim.exception'] is None
        assert os.path.exists(
            os.path.join(
                result['config']['meta.sim.workspace'],
                result['config']['sim.result.file'],
            )
        )

    with open(os.path.join(config['sim.workspace'], 'cookie.txt'), 'w') as f:
        f.write('hi')

    factors = [(['sim.seed'], [[1], [2]])]
    results = simulate_factors(config, factors, TopTest)
    assert len(results) == 2
    assert os.path.isdir(config['sim.workspace'])
    for result in results:
        assert result['sim.exception'] is None
        assert os.path.exists(
            os.path.join(
                result['config']['meta.sim.workspace'],
                result['config']['sim.result.file'],
            )
        )

    assert not os.path.exists(os.path.join(config['sim.workspace'], 'cookie.txt'))
    assert set(os.listdir(config['sim.workspace'])) == set(['0', '1'])


def test_progress_enabled(config):
    config['sim.progress.enable'] = True
    result = simulate(config, TopTest)
    assert result['sim.exception'] is None
    assert result['sim.now'] == 1
    assert result['sim.time'] == 1e-6
    assert result['sim.runtime'] > 0
    assert os.path.exists(
        os.path.join(config['sim.workspace'], config['sim.result.file'])
    )


@pytest.mark.parametrize('max_width', [0, 1])
def test_many_progress_enabled(config, max_width):
    config['sim.progress.enable'] = True
    config['sim.progress.max_width'] = max_width
    factors = [(['sim.seed'], [[1], [2], [3]])]
    results = simulate_factors(config, factors, TopTest)
    for result in results:
        assert result['sim.exception'] is None
        assert result['sim.now'] == 1
        assert result['sim.time'] == 1e-6
        assert result['sim.runtime'] > 0
        assert os.path.exists(
            os.path.join(
                result['config']['meta.sim.workspace'],
                result['config']['sim.result.file'],
            )
        )


def test_many_progress_no_pbar(config, capsys, no_progressbar):
    config['sim.progress.enable'] = True
    config['sim.duration'] = '10 us'
    factors = [(['sim.seed'], [[1], [2], [3]])]
    with capsys.disabled():
        simulate_factors(config, factors, TopTest)


def test_many_progress_no_colorama(config, capsys, no_colorama):
    config['sim.progress.enable'] = True
    factors = [(['sim.seed'], [[1], [2], [3]])]
    with capsys.disabled():
        simulate_factors(config, factors, TopTest)


def test_workspace_env_init(config):
    class TestEnvironment(SimEnvironment):
        def __init__(self, config):
            super().__init__(config)
            assert os.path.split(os.getcwd())[-1] == config['sim.workspace']

    workspace = config['sim.workspace']
    assert not os.path.exists(workspace)
    simulate(config, TopTest, TestEnvironment)
    assert os.path.exists(workspace)


def test_workspace_no_overwrite(config):
    workspace = config['sim.workspace']

    config['sim.workspace.overwrite'] = True
    config['sim.result.file'] = 'first-result.yaml'
    assert not os.path.exists(workspace)
    simulate(config, TopTest)
    assert os.path.exists(os.path.join(workspace, 'first-result.yaml'))

    config['sim.workspace.overwrite'] = False
    config['sim.result.file'] = 'second-result.yaml'
    simulate(config, TopTest)
    assert os.path.exists(os.path.join(workspace, 'first-result.yaml'))
    assert os.path.exists(os.path.join(workspace, 'second-result.yaml'))

    config['sim.workspace.overwrite'] = True
    config['sim.result.file'] = 'third-result.yaml'
    simulate(config, TopTest)
    assert not os.path.exists(os.path.join(workspace, 'first-result.yaml'))
    assert not os.path.exists(os.path.join(workspace, 'second-result.yaml'))
    assert os.path.exists(os.path.join(workspace, 'third-result.yaml'))


def test_workspace_is_curdir(config):
    config['sim.workspace'] = '.'
    config['sim.workspace.overwrite'] = True
    config['sim.result.file'] = 'first-result.yaml'
    simulate(config, TopTest)
    assert os.path.exists('first-result.yaml')

    config['sim.workspace.overwrite'] = True
    config['sim.result.file'] = 'second-result.yaml'
    simulate(config, TopTest)
    # '.' is not supposed to be overwritten
    assert os.path.exists('first-result.yaml')
    assert os.path.exists('second-result.yaml')


def test_many_with_duplicate_workspace(config):
    configs = [config.copy() for _ in range(2)]
    configs[0]['sim.workspace'] = os.path.join('tmp', os.pardir, 'workspace')
    configs[1]['sim.workspace'] = 'workspace'
    with pytest.raises(ValueError):
        simulate_many(configs, TopTest)


def test_many_user_jobs(config):
    simulate_many([config], TopTest, jobs=1)


def test_many_invalid_jobs(config):
    with pytest.raises(ValueError):
        simulate_many([config], TopTest, jobs=0)


def test_sim_time(config):
    config['sim.timescale'] = '10 ms'
    config['sim.duration'] = '995 ms'
    result = simulate(config, TopTest)
    assert result['sim.time'] == 0.995
    assert result['sim.now'] == 99.5
    assert result['time_100ps'] == 9950000000


def test_sim_time_non_default_t(config):
    config['sim.timescale'] = '1 ms'
    env = SimEnvironment(config)
    assert env.time(1000, 's') == 1
    assert env.time(1, 'ms') == 1
    assert env.time(t=500) == 0.5


@pytest.mark.parametrize('progress_enable', [True, False])
def test_sim_until(config, progress_enable):
    class TestEnvironment(SimEnvironment):
        def __init__(self, config):
            super().__init__(config)
            self.until = SimStopEvent(self)

    config['sim.progress.enable'] = progress_enable
    config['test.until_delay'] = 0
    result = simulate(config, TopTest, TestEnvironment)
    assert result['sim.now'] == 0.50

    config['test.until_delay'] = 0.25
    result = simulate(config, TopTest, TestEnvironment)
    assert result['sim.now'] == 0.75


def test_sim_json_result(config):
    config['sim.result.file'] = 'result.json'
    result = simulate(config, TopTest)
    workspace = config['sim.workspace']
    with open(os.path.join(workspace, config['sim.result.file'])) as f:
        assert json.load(f) == result


@pytest.mark.parametrize(
    'ext, parser',
    [
        ('yaml', yaml.safe_load),
        ('yml', yaml.safe_load),
        ('json', json.load),
        ('py', lambda f: eval(f.read())),
    ],
)
def test_sim_result_format(config, ext, parser):
    config['sim.result.file'] = 'result.' + ext
    config['sim.config.file'] = 'config.' + ext
    result = simulate(config, TopTest)
    workspace = config['sim.workspace']
    with open(os.path.join(workspace, config['sim.result.file'])) as f:
        assert parser(f) == result
    with open(os.path.join(workspace, config['sim.config.file'])) as f:
        assert parser(f) == config


def test_sim_invalid_result_format(config):
    config['sim.result.file'] = 'result.bogus'
    with pytest.raises(ValueError):
        simulate(config, TopTest)

    result = simulate(config, TopTest, reraise=False)
    assert result['sim.exception'] is not None



from typing import Optional, Tuple, Union
import re

_unit_map = {'s': 1e0, 'ms': 1e3, 'us': 1e6, 'ns': 1e9, 'ps': 1e12, 'fs': 1e15}

_num_re = r'[-+]? (?: \d*\.\d+ | \d+\.?\d* ) (?: [eE] [-+]? \d+)?'

_timescale_re = re.compile(rf'(?P<num>{_num_re})?\s?(?P<unit> [fpnum]? s)?', re.VERBOSE)

TimeValue = Tuple[Union[int, float], str]


def parse_time(time_str: str, default_unit: Optional[str] = None) -> TimeValue:
    """Parse a string containing a time magnitude and optional unit.

    :param str time_str: Time string to parse.
    :param str default_unit:
        Default time unit to apply if unit is not present in `time_str`. The
        default unit is only applied if `time_str` does not specify a unit.
    :returns:
        `(magnitude, unit)` tuple where magnitude is numeric (int or float) and
        the unit string is one of "s", "ms", "us", "ns", "ps", or "fs".
    :raises ValueError:
        If the string cannot be parsed or is missing a unit specifier and no
        `default_unit` is specified.

    """
    match = _timescale_re.match(time_str)
    if not match or not time_str:
        raise ValueError(f'Invalid timescale string "{time_str}"')
    if match.group('num'):
        num_str = match.group('num')
        num: Union[int, float]
        try:
            num = int(num_str)
        except ValueError:
            num = float(num_str)
    else:
        num = 1

    if match.group('unit'):
        unit = match.group('unit')
    else:
        if default_unit:
            unit = default_unit
        else:
            raise ValueError('No unit specified')
    return num, unit


def scale_time(from_time: TimeValue, to_time: TimeValue) -> Union[int, float]:
    """Scale time values.

    :param tuple from_time: `(magnitude, unit)` tuple to be scaled.
    :param tuple to_time: `(magnitude, unit)` tuple to scale to.
    :returns: Numeric scale factor relating `from_time` to `to_time`.

    """
    from_t, from_u = from_time
    to_t, to_u = to_time
    from_scale = _unit_map[from_u]
    to_scale = _unit_map[to_u]

    scaled = (to_scale / from_scale * from_t) / to_t

    if scaled % 1.0 == 0.0:
        return int(scaled)
    else:
        return scaled



from pytest import raises
import pytest

from desmod.pool import Pool, PriorityPool


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool(env, PoolClass):
    pool = PoolClass(env, capacity=2)

    def producer(amount, wait):
        yield env.timeout(wait)
        yield pool.put(amount)

    def consumer(expected_amount, wait):
        yield env.timeout(wait)
        msg = yield pool.get(expected_amount)
        assert msg == expected_amount

    env.process(producer(1, 0))
    env.process(producer(2, 1))
    env.process(consumer(1, 0))
    env.process(consumer(2, 1))
    env.process(consumer(2, 2))
    env.process(producer(1, 2))
    env.process(producer(1, 3))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool2(env, PoolClass):
    pool = PoolClass(env, capacity=2)

    def proc(env, pool):
        assert pool.is_empty
        assert env.now == 0

        yield env.timeout(1)

        when_full = pool.when_full()
        assert not when_full.triggered

        when_any = pool.when_any()
        assert not when_any.triggered

        with pool.when_not_full() as when_not_full:
            yield when_not_full
            assert when_not_full.triggered

        with raises(ValueError):
            pool.put(pool.capacity + 1)

        with raises(ValueError):
            pool.get(pool.capacity + 1)

        get_two = pool.get(2)
        assert not get_two.triggered

        put_one = pool.put(1)
        assert put_one.triggered

        assert not when_any.triggered
        assert not get_two.triggered
        assert not when_full.triggered
        assert pool.level == 1

        yield put_one
        assert when_any.triggered

        yield env.timeout(1)

        with pool.when_full() as when_full2:
            assert not when_full2.triggered

        put_one = pool.put(1)
        assert put_one.triggered
        assert not when_full.triggered

        yield put_one

        assert when_full.triggered
        assert get_two.triggered
        assert pool.level == 0

        yield pool.put(2)

        when_not_full = pool.when_not_full()
        assert not when_not_full.triggered

        with pool.when_any() as when_any2:
            yield when_any2
            assert when_any2.triggered

        yield pool.get(1)

        assert when_not_full.triggered

    env.process(proc(env, pool))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool_float(env, PoolClass):
    pool = PoolClass(env, capacity=3.0)

    def proc(env, pool):
        assert pool.is_empty

        when_full = pool.when_full()
        assert not when_full.triggered
        when_any = pool.when_any()
        assert not when_any.triggered

        get_half = pool.get(0.5)
        assert not get_half.triggered
        put_three = pool.put(3)
        assert put_three.triggered
        yield put_three
        assert pool.level == 2.5
        assert get_half.triggered

        with raises(AssertionError):
            when_not_full = pool.when_not_full()
        when_not_full = pool.when_not_full(epsilon=0.01)
        assert when_not_full.triggered

        put_half = pool.put(0.5)
        assert put_half.triggered
        yield put_half

        when_not_full = pool.when_not_full(epsilon=0.01)
        assert not when_not_full.triggered

    env.process(proc(env, pool))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool_overflow(env, PoolClass):
    pool = PoolClass(env, capacity=5, hard_cap=True)

    def producer(env):
        yield env.timeout(1)
        yield pool.put(1)
        yield pool.put(3)
        assert pool.remaining == 1
        with raises(OverflowError):
            yield pool.put(2)

    env.process(producer(env))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool_put_zero(env, PoolClass):
    pool = PoolClass(env, capacity=5, hard_cap=True)

    def producer(env):
        with raises(ValueError):
            yield pool.put(0)

    env.process(producer(env))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool_get_zero(env, PoolClass):
    pool = PoolClass(env, capacity=5, hard_cap=True)

    def consumer(env):
        with raises(ValueError):
            yield pool.get(0)

    env.process(consumer(env))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool_get_too_many(env, PoolClass):
    def producer(env, pool):
        yield pool.put(1)
        yield env.timeout(1)
        yield pool.put(1)

    def consumer(env, pool):
        amount = yield pool.get(1)
        assert amount == 1
        with raises(ValueError):
            yield pool.get(pool.capacity + 1)

    pool = PoolClass(env, capacity=6, name='foo')
    env.process(producer(env, pool))
    env.process(consumer(env, pool))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool_put_too_many(env, PoolClass):
    pool = PoolClass(env, capacity=6)

    def proc(env):
        with raises(ValueError):
            yield pool.put(pool.capacity + 1)

    env.process(proc(env))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool_cancel(env, PoolClass):
    def proc(env, pool):
        get_ev = pool.get(2)
        full_ev = pool.when_full()
        any_ev = pool.when_any()
        empty_ev = pool.when_empty()

        assert not any_ev.triggered
        assert empty_ev.triggered

        yield env.timeout(1)

        any_ev.cancel()

        with pool.put(1) as put_ev:
            yield put_ev

        assert not get_ev.triggered
        assert not any_ev.triggered

        with pool.when_empty() as empty_ev:
            assert not empty_ev.triggered

        get_ev.cancel()
        full_ev.cancel()

        yield pool.put(1)

        assert not get_ev.triggered
        assert pool.is_full
        assert not full_ev.triggered

        put_ev = pool.put(1)
        assert not put_ev.triggered

        yield env.timeout(1)
        put_ev.cancel()

        with pool.get(1) as get_ev2:
            yield get_ev2
        assert not put_ev.triggered

    env.process(proc(env, PoolClass(env, capacity=2)))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool_when_at_most(env, PoolClass):
    def proc(env, pool):
        yield pool.put(3)
        at_most = {}
        at_most[0] = pool.when_at_most(0)
        at_most[3] = pool.when_at_most(3)
        at_most[1] = pool.when_at_most(1)
        at_most[2] = pool.when_at_most(2)
        assert not at_most[0].triggered
        assert not at_most[1].triggered
        assert not at_most[2].triggered
        assert at_most[3].triggered

        yield pool.get(1)
        assert pool.level == 2
        assert not at_most[0].triggered
        assert not at_most[1].triggered
        assert at_most[2].triggered

        yield pool.get(1)
        assert pool.level == 1
        assert not at_most[0].triggered
        assert at_most[1].triggered

        yield pool.get(1)
        assert pool.level == 0
        assert at_most[0].triggered

    env.process(proc(env, PoolClass(env)))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_when_at_least(env, PoolClass):
    def proc(env, pool):
        at_least = {}
        at_least[3] = pool.when_at_least(3)
        at_least[0] = pool.when_at_least(0)
        at_least[2] = pool.when_at_least(2)
        at_least[1] = pool.when_at_least(1)
        assert at_least[0].triggered
        assert not at_least[1].triggered
        assert not at_least[2].triggered
        assert not at_least[3].triggered

        yield pool.put(1)
        assert at_least[1].triggered
        assert not at_least[2].triggered
        assert not at_least[3].triggered

        yield pool.get(1)
        assert not at_least[2].triggered
        assert not at_least[3].triggered

        yield pool.put(1)
        assert not at_least[2].triggered
        assert not at_least[3].triggered

        yield pool.put(1)
        assert at_least[2].triggered
        assert not at_least[3].triggered

        yield pool.put(1)
        assert at_least[3].triggered

    env.process(proc(env, PoolClass(env)))
    env.run()


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_pool_check_str(env, PoolClass):
    pool = PoolClass(env, name='bar', capacity=5)
    assert str(pool) == f"{PoolClass.__name__}(name='bar' level=0 capacity=5)"


def test_priority_pool_gets(env):
    pool = PriorityPool(env)

    def producer(env, pool):
        for _ in range(10):
            yield env.timeout(1)
            yield pool.put(1)

    def consumer(get_event):
        yield get_event

    get1_p1_a = env.process(consumer(pool.get(1, priority=1)))
    get1_p1_b = env.process(consumer(pool.get(1, priority=1)))
    get5_p0 = env.process(consumer(pool.get(5, priority=0)))
    get4_p0 = env.process(consumer(pool.get(4, priority=0)))

    env.process(producer(env, pool))

    env.run(until=5.1)
    assert get5_p0.triggered
    assert not get4_p0.triggered
    assert not get1_p1_a.triggered
    assert not get1_p1_b.triggered

    env.run(until=9.1)
    assert get4_p0.triggered
    assert not get1_p1_a.triggered
    assert not get1_p1_b.triggered

    env.run(until=10.1)
    assert get1_p1_a.triggered
    assert not get1_p1_b.triggered


def test_priority_pool_puts(env):
    def proc(env, pool):
        put_ev = {}
        put_ev[2] = pool.put(1, priority=2)
        put_ev[0] = pool.put(1, priority=0)
        put_ev[1] = pool.put(1, priority=1)
        assert not put_ev[0].triggered
        assert not put_ev[1].triggered
        assert not put_ev[2].triggered

        yield pool.get(1)
        assert put_ev[0].triggered
        assert not put_ev[1].triggered
        assert not put_ev[2].triggered

        yield pool.get(1)
        assert put_ev[1].triggered
        assert not put_ev[2].triggered

        yield pool.get(1)
        assert put_ev[2].triggered

    env.process(proc(env, PriorityPool(env, capacity=2, init=2)))
    env.run()



"""Queue classes useful for modeling.

A queue may be used for inter-process message passing, resource pools,
event sequences, and many other modeling applications. The :class:`~Queue`
class implements a simulation-aware, general-purpose queue useful for these
modeling applications.

The :class:`~PriorityQueue` class is an alternative to :class:`~Queue` that
dequeues items in priority-order instead of :class:`Queue`'s FIFO discipline.

"""
from heapq import heapify, heappop, heappush
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Type,
    TypeVar,
    Union,
)

from simpy.core import BoundClass, Environment
from simpy.events import Event

EventCallback = Callable[[Event], None]


class QueuePutEvent(Event):
    callbacks: List[EventCallback]

    def __init__(self, queue: 'Queue[ItemType]', item: Any) -> None:
        super().__init__(queue.env)
        self.queue = queue
        self.item = item
        queue._put_waiters.append(self)
        self.callbacks.extend([queue._trigger_when_at_least, queue._trigger_get])
        queue._trigger_put()

    def __enter__(self) -> 'QueuePutEvent':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.cancel()
        return None

    def cancel(self) -> None:
        if not self.triggered:
            self.queue._put_waiters.remove(self)
            self.callbacks = None  # type: ignore[assignment] # noqa: F821


class QueueGetEvent(Event):
    callbacks: List[EventCallback]

    def __init__(self, queue: 'Queue[ItemType]') -> None:
        super().__init__(queue.env)
        self.queue = queue
        queue._get_waiters.append(self)
        self.callbacks.extend([queue._trigger_when_at_most, queue._trigger_put])
        queue._trigger_get()

    def __enter__(self) -> 'QueueGetEvent':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.cancel()
        return None

    def cancel(self) -> None:
        if not self.triggered:
            self.queue._get_waiters.remove(self)
            self.callbacks = None  # type: ignore[assignment] # noqa: F821


class QueueWhenAtMostEvent(Event):
    def __init__(self, queue: 'Queue[ItemType]', num_items: Union[int, float]) -> None:
        super().__init__(queue.env)
        self.queue = queue
        self.num_items = num_items
        heappush(queue._at_most_waiters, self)
        queue._trigger_when_at_most()

    def __lt__(self, other: 'QueueWhenAtMostEvent') -> bool:
        return self.num_items > other.num_items

    def __enter__(self) -> 'QueueWhenAtMostEvent':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.cancel()
        return None

    def cancel(self) -> None:
        if not self.triggered:
            self.queue._at_most_waiters.remove(self)
            heapify(self.queue._at_most_waiters)
            self.callbacks = None  # type: ignore[assignment] # noqa: F821


class QueueWhenAtLeastEvent(Event):
    def __init__(self, queue: 'Queue[ItemType]', num_items: Union[int, float]) -> None:
        super().__init__(queue.env)
        self.queue = queue
        self.num_items = num_items
        heappush(queue._at_least_waiters, self)
        queue._trigger_when_at_least()

    def __lt__(self, other: 'QueueWhenAtLeastEvent') -> bool:
        return self.num_items < other.num_items

    def __enter__(self) -> 'QueueWhenAtLeastEvent':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.cancel()
        return None

    def cancel(self) -> None:
        if not self.triggered:
            self.queue._at_least_waiters.remove(self)
            heapify(self.queue._at_least_waiters)
            self.callbacks = None  # type: ignore[assignment] # noqa: F821


class QueueWhenAnyEvent(QueueWhenAtLeastEvent):
    def __init__(self, queue: 'Queue[ItemType]') -> None:
        super().__init__(queue, num_items=1)


class QueueWhenFullEvent(QueueWhenAtLeastEvent):
    def __init__(self, queue: 'Queue[ItemType]') -> None:
        super().__init__(queue, num_items=queue.capacity)


class QueueWhenNotFullEvent(QueueWhenAtMostEvent):
    def __init__(self, queue: 'Queue[ItemType]') -> None:
        super().__init__(queue, num_items=queue.capacity - 1)


class QueueWhenEmptyEvent(QueueWhenAtMostEvent):
    def __init__(self, queue: 'Queue[ItemType]') -> None:
        super().__init__(queue, num_items=0)


ItemType = TypeVar('ItemType')


class Queue(Generic[ItemType]):
    """Simulation queue of arbitrary items.

    `Queue` is similar to :class:`simpy.Store`. It provides a simulation-aware
    first-in first-out (FIFO) queue useful for passing messages between
    simulation processes or managing a pool of objects needed by multiple
    processes.

    Items are enqueued and dequeued using :meth:`put()` and :meth:`get()`.

    :param env: Simulation environment.
    :param capacity: Capacity of the queue; infinite by default.
    :param hard_cap:
        If specified, the queue overflows when the `capacity` is reached.
    :param items: Optional sequence of items to pre-populate the queue.
    :param name: Optional name to associate with the queue.

    """

    def __init__(
        self,
        env: Environment,
        capacity: Union[int, float] = float('inf'),
        hard_cap: bool = False,
        items: Iterable[ItemType] = (),
        name: Optional[str] = None,
    ) -> None:
        self.env = env
        #: Capacity of the queue (maximum number of items).
        self.capacity = capacity
        self._hard_cap = hard_cap
        self.items: List[ItemType] = list(items)
        self.name = name
        self._put_waiters: List[QueuePutEvent] = []
        self._get_waiters: List[QueueGetEvent] = []
        self._at_most_waiters: List[QueueWhenAtMostEvent] = []
        self._at_least_waiters: List[QueueWhenAtLeastEvent] = []
        self._put_hook: Optional[Callable[[], Any]] = None
        self._get_hook: Optional[Callable[[], Any]] = None
        BoundClass.bind_early(self)

    @property
    def size(self) -> int:
        """Number of items in queue."""
        return len(self.items)

    @property
    def remaining(self) -> Union[int, float]:
        """Remaining queue capacity."""
        return self.capacity - len(self.items)

    @property
    def is_empty(self) -> bool:
        """Indicates whether the queue is empty."""
        return not self.items

    @property
    def is_full(self) -> bool:
        """Indicates whether the queue is full."""
        return len(self.items) >= self.capacity

    def peek(self) -> ItemType:
        """Peek at the next item in the queue."""
        return self.items[0]

    if TYPE_CHECKING:

        def put(self, item: ItemType) -> QueuePutEvent:
            """Enqueue an item on the queue."""
            ...

        def get(self) -> QueueGetEvent:
            """Dequeue an item from the queue."""
            ...

        def when_at_least(self, num_items: int) -> QueueWhenAtLeastEvent:
            """Return an event triggered when the queue has at least n items."""
            ...

        def when_at_most(self, num_items: int) -> QueueWhenAtMostEvent:
            """Return an event triggered when the queue has at most n items."""
            ...

        def when_any(self) -> QueueWhenAnyEvent:
            """Return an event triggered when the queue is non-empty."""
            ...

        def when_full(self) -> QueueWhenFullEvent:
            """Return an event triggered when the queue becomes full."""
            ...

        def when_not_full(self) -> QueueWhenNotFullEvent:
            """Return an event triggered when the queue becomes not full."""
            ...

        def when_empty(self) -> QueueWhenEmptyEvent:
            """Return an event triggered when the queue becomes empty."""
            ...

    else:
        put = BoundClass(QueuePutEvent)
        get = BoundClass(QueueGetEvent)
        when_at_least = BoundClass(QueueWhenAtLeastEvent)
        when_at_most = BoundClass(QueueWhenAtMostEvent)
        when_any = BoundClass(QueueWhenAnyEvent)
        when_full = BoundClass(QueueWhenFullEvent)
        when_not_full = BoundClass(QueueWhenNotFullEvent)
        when_empty = BoundClass(QueueWhenEmptyEvent)

    def _enqueue_item(self, item: ItemType) -> None:
        self.items.append(item)

    def _dequeue_item(self) -> ItemType:
        return self.items.pop(0)

    def _trigger_put(self, _: Optional[Event] = None) -> None:
        while self._put_waiters:
            if len(self.items) < self.capacity:
                put_ev = self._put_waiters.pop(0)
                self._enqueue_item(put_ev.item)
                put_ev.succeed()
                if self._put_hook:
                    self._put_hook()
            elif self._hard_cap:
                raise OverflowError()
            else:
                break

    def _trigger_get(self, _: Optional[Event] = None) -> None:
        while self._get_waiters and self.items:
            get_ev = self._get_waiters.pop(0)
            item = self._dequeue_item()
            get_ev.succeed(item)
            if self._get_hook:
                self._get_hook()

    def _trigger_when_at_least(self, _: Optional[Event] = None) -> None:
        while (
            self._at_least_waiters and self.size >= self._at_least_waiters[0].num_items
        ):
            when_at_least_ev = heappop(self._at_least_waiters)
            when_at_least_ev.succeed()

    def _trigger_when_at_most(self, _: Optional[Event] = None) -> None:
        while self._at_most_waiters and self.size <= self._at_most_waiters[0].num_items:
            at_most_ev = heappop(self._at_most_waiters)
            at_most_ev.succeed()

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}('
            f'name={self.name!r} size={self.size} capacity={self.capacity})'
        )


class PriorityItem(NamedTuple):
    """Wrap items with explicit priority for use with :class:`~PriorityQueue`.

    :param priority:
        Orderable priority value. Smaller values are dequeued first.
    :param item:
        Arbitrary item. Only the `priority` is determines dequeue order, so the
        `item` itself does not have to be orderable.

    """

    priority: Any
    item: Any

    def __lt__(  # type: ignore[override] # noqa: F821
        self, other: 'PriorityItem'
    ) -> bool:
        return self.priority < other.priority


class PriorityQueue(Queue[ItemType]):
    """Specialized queue where items are dequeued in priority order.

    Items in `PriorityQueue` must be orderable (implement
    :meth:`~object.__lt__`). Unorderable items may be used with `PriorityQueue`
    by wrapping with :class:`~PriorityItem`.

    Items that evaluate less-than other items will be dequeued first.

    """

    def __init__(
        self,
        env: Environment,
        capacity: Union[int, float] = float('inf'),
        hard_cap: bool = False,
        items: Iterable[ItemType] = (),
        name: Optional[str] = None,
    ) -> None:
        super().__init__(env, capacity, hard_cap, items, name)
        heapify(self.items)

    def _enqueue_item(self, item: ItemType) -> None:
        heappush(self.items, item)

    def _dequeue_item(self) -> ItemType:
        return heappop(self.items)
